/*
*	trans: string translator
*	ver. 1.0
*	-Kenar- (Kenji Arisawa)
*	Email: arisawa@aichi-u.ac.jp
*/


#include <u.h>
#include <libc.h>
#include <bio.h>

#define	DBG	if(0)
/*	note: if(0) { ... }
*	if(0) strips the object codes from *.$O.
*	you need not remove debugging codes from the source to save object code size.
*/

#ifdef __APPLE__
#define UNIX
#endif

#ifdef unix
#define UNIX
/* stime conflict on linux */
#define stime s_time
#endif


void
usage(void)
{
	fprint(2,"usage: trans rule [file ....]\n");
	exits("usage");
}

#define ERR (-1)

int _ungetcount;
uchar _ungetbuf[512];

typedef struct Ttab Ttab;
struct Ttab {
char *frstr;
char *tostr;
int	frlen; /* sizeof frstr */
};

Biobuf bout;

typedef struct List List;

struct List{
	void *data;
	List *next;
};

List* Lnew(void *d)
{
	List *p;
	p = (List*) malloc(sizeof(List));
	p->next = 0;
	p->data = d;
	return p;
};

List* Lappend(List *p, void *d)
{
	List *q;
	q = (List*) malloc(sizeof(List));
	q->next = p->next;
	q->data = d;
	p->next = q;
	return q;
};

void Lfree(List *p)
{
	List *q;
	q = p->next;
	free(p);
	while(q != 0){
		p = q->next;
		free(q);
		q = p;
	}
}

#ifdef XXX
/*	this main() show the usage of
*	Lnew(), Lappend() and Lfree()
*	these are simple function for List	*/
void main(int argc, char *argv[])
{
	List *a, *p;

	USED(argc);
	USED(argv);

	a = Lnew("alice");
	p = Lappend(a,"bob");
	p = Lappend(p,"carol");

	for (p = a; p; p = p->next)
		print("%s\n", p->data);

	Lfree(a);
}
#endif


/*
 * Bgetch: extended Bgetc()
 */
int
Bgetch(Biobuf *bin)
{
	int c;
	if(_ungetcount > 0)
		return _ungetbuf[--_ungetcount];
	c = Bgetc(bin);
	if(c < 1)
		return ERR;
	return (unsigned)c&0xff;
}

/*
 * Bungetch: unget n chars.
 */
void
Bungetch(char *s, int n)
{
	while(n > 0)
		_ungetbuf[_ungetcount++] = s[--n];
}

enum{
	/* some rule file formats */
	rfmt = 0, /* raw utf-8 string, default */
	hfmt = 1, /* utf-8 in hex */
	ufmt = 2, /* unicode in hex */
};

void
Btrans(Biobuf *bop, Biobuf *bip, Ttab *ttab, int n, char *buf)
/*	bop: pointer to bout
*	bip: pointer to bin
*	ttab: table of rule
*	n: number of rules
*	buf: working buf, the size must be >= max len of ttab[n].frstr	*/
{
	int m,j,k,c;

	for(;;)
	{
Lstart:
		c = 0;
		m = 0; j = 0; k = n;
		while(j < k)
		{
			c = Bgetch(bip);
			DBG print("|%02x",c);
			if(c == ERR) break;
			while (j <= n && c > (uchar)ttab[j].frstr[m]) j++;
			while (k >= 0 && c < (uchar)ttab[k].frstr[m]) k--;
			buf[m++] = c;
		}
		DBG print("n,m,j,k:%d,%d,%d,%d,%c(%02x)\n", n,m,j,k,c,c);
		if(j == k)
		{
			while(m < ttab[j].frlen)
			{
				c = Bgetch(bip);
				DBG print("^m:%d,%02x(%02x)\n",m,c,(uchar)ttab[j].frstr[m]);
				if(c == ERR){
					if(m == 0)
						return;
					break;
				}
				buf[m] = c;
				if((uchar)ttab[j].frstr[m] != c) break;
				m++;
			}
			DBG print("m,j:%d,%d\n",m,j);
			if(m == ttab[j].frlen){ /* found */
				Bwrite(bop, ttab[j].tostr, strlen(ttab[j].tostr));
				goto Lstart;
			}
			if(m > 0){
				Bungetch(&buf[1],m);
			}
			Bwrite(bop, buf, 1);
			continue;
		}
		Bwrite(bop,buf,m);
		if(c == ERR || c == 0)
			return;
	}
}



/*
 * getch: read chars from buf
 */
int
getch(char *buf)
{
	int c;
	static char *b0;
	static char *p = nil;

	if(b0 != buf)
		p = b0 = buf;
	if(_ungetcount > 0)
		return _ungetbuf[--_ungetcount];
	c = *p++;
	if(c == 0)
		return ERR;
	return (unsigned)c&0xff;
}


/*
 * ungetch: unget n chars.
 */
void
ungetch(char *s, int n)
{
	while(n > 0)
		_ungetbuf[_ungetcount++] = s[--n];
}


char *
sputs(char *wp, char *ep, char *s)
/*	ep: end point	*/
{
	int n;

	n = strlen(s);
	if(wp + n > ep)
		n = ep - wp;
	memcpy(wp,s,n);
	wp += n;
	*wp = 0;
	return wp;
}

char *
etrans(char *result, char *ep, char *source, Ttab *ttab, int n, char *buf)
/*	result: buf to store the result
*	ep: size of result
*	source: source text
*	ttab: table of rule
*	n: number of rules
*	buf: working buf, the size must be >= max len of ttab[n].frstr	*/
{
	int m,j,k,c;
	char *wp = result;

	for(;;)
	{
Lstart:
		c = 0;
		m = 0; j = 0; k = n;
		while(j < k)
		{
			c = getch(source);
			DBG print("|%02x",c);
			if(c == ERR) break;
			while (j <= n && c > (uchar)ttab[j].frstr[m]) j++;
			while (k >= 0 && c < (uchar)ttab[k].frstr[m]) k--;
			buf[m++] = c;
		}
		DBG print("n,m,j,k:%d,%d,%d,%d,%c(%02x)\n", n,m,j,k,c,c);
		if(j == k)
		{
			while(m < ttab[j].frlen)
			{
				c = getch(source);
				DBG print("^m:%d,%02x(%02x)\n",m,c,(uchar)ttab[j].frstr[m]);
				if(c == ERR){
					if(m == 0)
						return wp;
					break;
				}
				buf[m] = c;
				if((uchar)ttab[j].frstr[m] != c) break;
				m++;
			}
			DBG print("m,j:%d,%d\n",m,j);
			if(m == ttab[j].frlen){ /* found */
				wp = sputs(wp,ep,ttab[j].tostr);
				goto Lstart;
			}
			if(m > 0){
				ungetch(&buf[1],m);
			}
			buf[1] = 0;
			wp = sputs(wp, ep, buf);
			continue;
		}
		buf[m] = 0;
		wp = sputs(wp,ep,buf);
		if(c == ERR || c == 0)
			return wp;
	}
}


/*
*	translate1 shows the usage example of Btrans()
*	Btrans() is designed for stream conversion
*	stdin to stdout
*	Binary data should be OK.
*/
void
translate1(int fd, Ttab *ttab)
{
	Biobuf bin;
	char *buf;
	int n,m;

	for(n = 0; ttab[n].frstr; n++);
	DBG print("%d\n",n);

	m = 0;
	for(n = 0; ttab[n].frstr; n++){
		DBG print("%s %s\n",ttab[n].frstr,ttab[n].tostr);
		if(ttab[n].frlen > m)
			m = ttab[n].frlen;
	}

	/* then m is max size of frlen */
	buf = malloc(m);
	n--;
	/*  then n is the number of rules (a pair of strings) */
	Binit(&bin, fd, OREAD);
	Btrans(&bout, &bin, ttab, n, buf);
	free(buf);
	Bflush(&bout);
}


/*
*	translate2 shows the usage example of etrans()
*	etrans() is designed for string to string conversion
*	these string must be utf-8.
*/
void
translate2(int fd, Ttab *ttab)
{
	Biobuf bin;
	char *buf;
	int n,m;
	char obuf[512], *ep, *wp;
	char *source;
	ep = obuf + 512;

	for(n = 0; ttab[n].frstr; n++);
	DBG print("%d\n",n);

	m = 0;
	for(n = 0; ttab[n].frstr; n++){
		DBG print("%s %s\n",ttab[n].frstr,ttab[n].tostr);
		if(ttab[n].frlen > m)
			m = ttab[n].frlen;
	}

	/* then m is max size of frlen */
	buf = malloc(m + 1); // need space for '\0'
	n--;
	/*  then n is the number of rules (a pair of strings) */
	Binit(&bin, fd, OREAD);
	while((source = Brdstr(&bin,'\n',1))){ /* assign = */
		DBG print("!%s\n",source);
		wp = etrans(obuf, ep, source, ttab, n, buf);
		if(wp == ep)
			fprint(2,"# output buffer full\n");
		*wp = 0; /* not needed. internally done by sputs() */
		Bprint(&bout,"%s\n",obuf);
	}
	free(buf);
	Bflush(&bout);
}

char *
tr(char *str, int fmt)
/*	tr translate various formatted data to utf-8 raw format.
*	str: input
*	fmt: format style
*	return: converted str.
*	NB: str will be scratched */
{
	char *s,*end;
	s = str;

	if(fmt == hfmt){
		int i,ch;
		for(i = 0;;){
			ch = strtol(s,&end,16);
			str[i++] = ch;
			if(*end == 0)
				break;
			s = end+1;
		}
		str[i] = 0;
		return str;
	}
	if(fmt == ufmt){
		int i,n;
		Rune rune;
		for(i = 0;;){
			rune = strtol(s,&end,16);
			/* typedef	ushort		Rune;	*/
			n = runetochar(&str[i],&rune);
			i += n;
			if(*end == 0)
				break;
			s = end+1;
		}
		str[i] = 0;
		return str;
	}
	return str;
}

/*	man strcat(2): 
*	The comparisons are made with unsigned bytes.	*/
int
compar(Ttab *x, Ttab *y)
{
	return strcmp(x->frstr,y->frstr);
}


/*
*	tabread: read translation data from rule file
*	lines beginning with "#" are comment
*	if the first line is comment and string "unicode" or "hex" is in it,
*	these are keywords for format of the file 	*/
Ttab *
tabread(char *path, int fmt)
{
	int n,i,m, fd;
	Biobuf bin;
	Ttab *transtab, *tp;
	char *p;
	List *q, *q0;
	char *args[3];
	long ch;

	/*
	*	trans data init
	*/
	fd = open(path, OREAD);
	if(fd < 0)
		sysfatal("%s not open %r", path);
	if(Binit(&bin, fd, OREAD) < 0)
		sysfatal("error in Binit: %r");
	ch = Bgetc(&bin);
	if(ch == '#'){
		p = Brdstr(&bin, '\n',1);
		if(cistrstr(p,"unicode"))
			fmt = ufmt;
		else
		if(cistrstr(p,"hex"))
			fmt = hfmt;
	}
	else
		Bungetc(&bin);

	q0 = q = Lnew(nil);
	n = 0;
	while((p = Brdstr(&bin, '\n', 1))){ /* assign = */
		if(p[0] == '#')
			continue;
		DBG print("%s\n",p);
		q = Lappend(q, strdup(p));
		n++;
	}
	close(fd);
	if(n == 0)
		sysfatal("# no rule");
	DBG print("done\n");

	transtab = calloc(n + 1, sizeof(Ttab));
	/* need room for nil */
	q0 = q0->next;

	for(i = 0, q = q0; q; q = q->next){
		m = tokenize(q->data,args, 3);
		if(m < 2)
			continue;
		tp = &transtab[i];
		tp->frstr = tr(args[0],fmt);
		tp->tostr = tr(args[1],fmt);
		tp->frlen = strlen(tp->frstr);
		if(tp->frlen == 0)
			sysfatal("# null string in the first field\n");
		DBG print("#%s %s\n", args[0], args[1]);//DBG
		i++;
	}
	transtab[i].frstr = nil;
	transtab[i].tostr = nil;
	transtab[i].frlen = -1;

	/* we should sort transtab by frstr */


    /*
	* Plan9
	* void qsort(void *base, long nel, long width,
    *   int (*compar)(void*, void*));
	*
	* UNIX
    * void qsort(void *base, size_t nel, size_t width,
	*   int (*compar)(const void *, const void *));
	*/
	n = i; /* num of element */

#ifdef UNIX
	qsort(transtab, n, sizeof(Ttab), (int (*)(const void*, const void*))compar);
#else
	qsort(transtab, n, sizeof(Ttab), compar);
#endif

	/* DBG */
	DBG for(i = 0; i < n; i++)
		print("# %02x %s %s\n", (uchar)transtab[i].frstr[0], transtab[i].frstr , transtab[i].tostr );
	return transtab;
}

/* 	which translate do you want to use?
*	translate1 or translate2
*	translate1 is better for the tool that handle stream
*	translate2 here is only for debugging purpose
*	we need the codes in translate2 for u9fs */
#define translate translate1

void
main(int argc, char *argv[])
{
	int fd, fmt = 0;
	Ttab *transtab;

	ARGBEGIN{
	default: usage();
	}ARGEND

	if(*argv == nil)
		usage();

	Binit(&bout, 1, OWRITE);

	transtab = tabread(*argv++, fmt);

	if(*argv == nil){
		translate(0,transtab);
		exits(nil);
	}

	while(*argv){
		fd = open(*argv++,OREAD);
		translate(fd,transtab);
	}
	exits(nil);
}
